# -*- coding: utf-8 -*-
"""
Created on Tue Oct  9 11:22:22 2018

@author: michaelek
"""
from time import time
import numpy as np
import pandas as pd
import xarray as xr
from hydrointerp.io.raster import save_geotiff
from hydrointerp.util import grp_ts_agg
from nzmetservice import select_bounds, to_df
from scipy import ndimage

pd.options.display.max_columns = 10

#####################################
### Parameters

min_lat = -47
max_lat = -40
min_lon = 166
max_lon = 175
nc1 = r'N:\met_service\forecasts\wrf_hourly_precip_nz4kmN-NCEP_2018090918.nc'
nc2 = r'N:\nasa\precip\gpm_3IMERGHHL\gpm_3IMERGHHL_v07_20190101-20190108.nc4'


from_crs = 4326
to_crs = 2193
grid_res = 1000
interp_fun = 'cubic'

time_col = 'time'
x_col = 'longitude'
y_col = 'latitude'
data_col = 'precip_rate'
interp_fun = 'cubic'
digits = 2

ts_resample_code = '4H'

point_path = r'N:\met_service\point_test1.shp'
point_site_col = 'site_id'
site_test = 5


####################################
### Import

ms1 = select_bounds(nc1, min_lat, max_lat, min_lon, max_lon)

ms_df = to_df(ms1, True).dropna().reset_index()
ms_df1 = ms_df[(ms_df.time <= '2018-09-24 12:00') & (ms_df.time >= '2018-09-24')]

### Resample in time
ms_df2 = grp_ts_agg(ms_df1, ['longitude', 'latitude'], 'time', ts_resample_code, 'left', 'right')[data_col].sum().reset_index()


### Interp

new_df = interp_to_grid(ms_df2, 'time', 'longitude', 'latitude', 'precip_rate', grid_res, from_crs, to_crs, interp_fun)

new_ds = interp_to_grid(ms_df2, 'time', 'longitude', 'latitude', 'precip_rate', grid_res, from_crs, to_crs, interp_fun, output='xarray')

new_points = interp_to_points(ms_df2, 'time', 'longitude', 'latitude', 'precip_rate', point_shp, point_site_col, from_crs)


###################################
### Testing

nc2 = r'N:\nasa\precip\gpm_3IMERGHHL_v07_20190101-20190106.nc4'
time_name = 'time'
x_name = 'lon'
y_name = 'lat'
data_name = 'precipitationCal'
from_crs = 4326
to_crs = 2193
grid_res = 1000
bbox=None
order=1
extrapolation='constant'
cval=np.nan
digits = 2
min_lat = -48
max_lat = -41
min_lon = 170
max_lon = 178
min_val=0
method='linear'
point_data = r'N:\met_service\point_test1.shp'

grid1 = xr.open_dataset(nc2)

grid = grid1.where((grid1[y_name] >= min_lat) & (grid1[y_name] <= max_lat) & (grid1[x_name] >= min_lon) & (grid1[x_name] <= max_lon), drop=True).copy()
grid1.close()

grid2d = grid.isel(time=280)[data_name]

grid2d.plot.pcolormesh(x='lon', y='lat')

start1 = time()
interp1 = grid_to_grid(grid, time_name, x_name, y_name, data_name, grid_res, from_crs, to_crs, bbox, order, extrapolation, cval, digits, min_val)
end1 = time()
end1 - start1

output2d = interp1.isel(time=280)[data_name]

output2d.plot.pcolormesh(x='x', y='y')

df = grid.to_dataframe().reset_index()

x = da1[x_name].values
y = da1[y_name].values

xy_orig_pts = np.dstack(np.meshgrid(x, y)).reshape(-1, 2)

da2 = da1.transpose(x_name, y_name, time_name)
ar1 = da2.values
ar1[:2] = np.nan

np.nan_to_num(ar1, False)

points_x, points_y = np.broadcast_arrays(xinterp.reshape(-1,1), yinterp)
coord = np.vstack((points_x.flatten()*(len(xgrid)-1), points_y.flatten()*(len(ygrid)-1)))


grouped = df1.groupby([x_name, y_name, time_name])[data_name].mean()

grouped = df1.set_index([x_name, y_name, time_name])[data_name]

# create an empty array of NaN of the right dimensions
shape = tuple(len(i) for i in grouped.index.levels)
arr = np.full(shape, np.nan)

# fill it using Numpy's advanced indexing
arr[tuple(grouped.index.labels)] = grouped.values.flat


df = pd.DataFrame({'x': [1, 2, 1, 3, 1, 2, 3, 1, 2],
                   'y': [1, 1, 2, 2, 1, 1, 1, 2, 2],
                   'z': [1, 1, 1, 1, 2, 2, 2, 2, 2],
                   'value': [1, 2, 3, 4, 5, 6, 7, 8, 9]})

grouped = df.groupby(['z', 'y', 'x'])['value'].mean()


a = np.arange(12.).reshape((4, 3))




dx, dy = 0.4, 0.4
xmax, ymax = 2, 4
x = np.arange(-xmax, xmax, dx)
y = np.arange(-ymax, ymax, dy)
X, Y = np.meshgrid(x, y)
Z = np.exp(-(2*X)**2 - (Y/2)**2)


interp2b = grid_to_points(grid, time_name, x_name, y_name, data_name, point_data, from_crs, to_crs, order, extrapolation, fill_val, digits, min_val)

interp2

d2 = df[df.time == pd.Timestamp('2019-01-06T20:29:59.999000')].copy()

interp2c = points_to_grid(d2, time_name, x_name, y_name, data_name, grid_res, from_crs, to_crs, bbox, method, fill_val, digits, min_val)

output2d = interp2c.isel(time=0)[data_name]

output2d.plot.pcolormesh(x='x', y='y')




grid1 = xr.open_dataset(nc1)
grid = grid1.copy()
grid1.close()


nc2 = r'N:\nasa\precip\gpm_3IMERGHH\gpm_3IMERGHH_v07_20170101-20171231.nc4'
nc2 = r'N:\nasa\precip\gpm_3IMERGHH\gpm_3IMERGHH_v07_20180101-20180701.nc4'

grid1 = xr.open_dataset(nc2)
da = grid1[data_name].copy()
grid1.close()

da1 = da * 0.5

da2 = da1.resample(time='D', closed='right', label='left').sum('time')

max_day = da2.groupby('time').sum().to_dataframe().sort_values('precipitationCal')

da3 = da2.loc['2018-01-15':'2018-01-22']

#grid2d.plot.pcolormesh(x='lon', y='lat')

plt1 = da3.plot(x='lon', y='lat', col='time', col_wrap=4)

save1 = r'E:\ecan\git\hydrointerp\hydrointerp\datasets\nasa_gpm_2017-07-20.nc'
save2 = r'E:\ecan\git\hydrointerp\hydrointerp\datasets\nasa_gpm_2017-07-20.tif'

da4 = da1.loc['2017-07-20':'2017-07-21 00:00']
da5 = da4.resample(time='H', closed='right', label='right').sum('time')
da5.attrs = grid1[data_name].attrs

da5.to_dataset().to_netcdf(save1)

#da5 = xr.open_dataset(save1)

df5 = da5.resample(time='D', closed='right', label='left').sum('time').to_dataframe().reset_index()

save_geotiff(df5, 4326, 'precipitationCal', 'lon', 'lat', export_path=save2)


a = np.arange(50, step=2).reshape((5,5))
a[2,2] = 2

ndimage.gaussian_filter(a, sigma=1, order=0)
ndimage.gaussian_gradient_magnitude(a, sigma=1)

ndimage.median_filter(a, size=3)
ndimage.percentile_filter(a, percentile=100, size=2)
ndimage.uniform_filter(a, size=3)



date0 = '2018-01-17'

date1 = pd.Timestamp(date0)
date2 = date1 + pd.DateOffset(days=1)

da4 = da1.loc[date1:date2]
da5 = da4.resample(time='D', closed='right', label='left').sum('time')
da6 = da5.copy()
da6['time'] = da6['time'].to_series() + pd.DateOffset(hours=1)

#da6.data = ndimage.percentile_filter(da5, percentile=100, size=2)
da6.data = ndimage.median_filter(da5, size=2)

da7 = da5.combine_first(da6)

plt1 = da7.plot(x='lon', y='lat', col='time', col_wrap=2)

save2 = r'E:\ecan\git\hydrointerp\hydrointerp\datasets\nasa_gpm_2018-01-18.tif'
save_geotiff(da6.to_dataframe().reset_index(), 4326, 'precipitationCal', 'lon', 'lat', export_path=save2)


from nasadap import Nasa

product = '3IMERGDF'
product = '3IMERGHH'
#product2c = '3IMERGHHE'
product = '3B42'
username = 'Dryden'
password = 'NasaData4me'
from_date = '2018-04-15'
to_date = '2018-04-17'
mission = 'gpm'
mission = 'trmm'
dataset_type = ['precipitationCal', 'randomError', 'precipitationQualityIndex']
dataset_type = ['precipitation']
min_lat=-49
max_lat=-33
min_lon=165
max_lon=180
cache_dir = r'\\fs02\GroundWaterMetData$\nasa\cache\nz'

ge = Nasa(username, password, mission, cache_dir)
ds2 = ge.get_data(product, dataset_type, from_date, to_date, min_lat, max_lat, min_lon, max_lon, 60)
ds3 = ds2.copy().load()

ds3['time'] = ds2.time.to_index() + pd.DateOffset(hours=12)
del ds2

da5 = ds3['precipitationCal'].resample(time='D', closed='right', label='left').sum('time').isel(time=slice(1, 4))
ea1 = ds3['randomError'].resample(time='D', closed='right', label='left').mean('time').isel(time=slice(1, 4))
qa1 = ds3['precipitationQualityIndex'].resample(time='D', closed='right', label='left').mean('time').isel(time=slice(1, 4))

ea2 = ea1/da5

plt1 = da5.plot(x='lon', y='lat', col='time', col_wrap=3)

da6 = da5.where(qa1 > 0.4).fillna(400)
da6.name = 'precip'

plt2 = da6.plot(x='lon', y='lat', col='time', col_wrap=3)


bad_spot1 = [168.75, -44.45]

day1 = '2018-01-17'
qa1.sel(time='2018-01-17', lat=-44.45, lon=168.75)
qa1.loc['2018-01-17'].sel(lon=168.75)

ea1.loc['2018-01-17'].sel(lon=168.75)

(qa1 < 0.26).sum(['lat', 'lon'])

qa2 = da5.where(qa1 > 0.26)
qa2.name = 'precip'

df1 = qa2.to_dataframe().reset_index()

### A 'precipitationQualityIndex' of > 0.4 is the ideal setting for GPM data


da5 = ds3['precipitation'].resample(time='D', closed='right', label='left').sum('time').isel(time=slice(1, 4))

plt1 = da5.plot(x='lon', y='lat', col='time', col_wrap=3)



da5 = ds3['precipitationCal'].resample(time='D', closed='right', label='left').sum('time').isel(time=slice(1, 4))
qa1 = ds3['precipitationQualityIndex'].resample(time='D', closed='right', label='left').mean('time').isel(time=slice(1, 4))

da6 = da5.where(qa1 > 0.4)
da6.name = 'precip'

ds7 = grid_interp_na(da6.to_dataset(), time_name, x_name, y_name, 'precip', min_val=0)

#da7 = ds7['precipitationCal'].resample(time='D', closed='right', label='left').sum('time').isel(time=slice(1, 4))

plt1 = da5.plot(x='lon', y='lat', col='time', col_wrap=3)
plt2 = ds7['precip'].plot(x='lon', y='lat', col='time', col_wrap=3)



























# -*- coding: utf-8 -*-
"""
Utility functions.
"""
import numpy as np
#from pycrs import parse
#
##########################################
#### Dictionaries for convert_crs
#
#proj4_netcdf_var = {'x_0': 'false_easting', 'y_0': 'false_northing', 'f': 'inverse_flattening',
#                    'lat_0': 'latitude_of_projection_origin',
#                    'lon_0': ('longitude_of_central_meridian', 'longitude_of_projection_origin'),
#                    'pm': 'longitude_of_prime_meridian',
#                    'k_0': ('scale_factor_at_central_meridian', 'scale_factor_at_projection_origin'),
#                    'a': 'semi_major_axis', 'b': 'semi_minor_axis', 'lat_1': 'standard_parallel',
#                    'proj': 'transform_name'}
#
#proj4_netcdf_name = {'aea': 'albers_conical_equal_area', 'tmerc': 'transverse_mercator',
#                     'aeqd': 'azimuthal_equidistant', 'laea': 'lambert_azimuthal_equal_area',
#                     'lcc': 'lambert_conformal_conic', 'cea': 'lambert_cylindrical_equal_area',
#                     'longlat': 'latitude_longitude', 'merc': 'mercator', 'ortho': 'orthographic',
#                     'ups': 'polar_stereographic', 'stere': 'stereographic', 'geos': 'vertical_perspective'}
#
#########################################
#### Functions


def grid_xy_to_map_coords(xy, digits=2, dtype=int, copy=True):
    """
    Convert an array of gridded x and y values to map_coordinates appropriate values.

    Parameters
    ----------
    xy : ndarray
        of shape (l, 2) where l is the number of location pairs. The first value of each pair is x and the second is y.
    digits : int
        The resolution of the data in digits.
    copy : bool
        Should it make a copy or modify the array in place?

    Returns
    -------
    ndarray
        Transposed shape from the input xy to an index array.
    dxy
        The grid interval.
    x_min
        The x value at index 0.
    y_min
        The y value at index 0.
    """
    if copy:
        xy1 = (xy.T * 10**digits).copy()
    else:
       xy1 = (xy.T * 10**digits)
    x_min, y_min = xy1.min(1)
    dxy = int(np.median(np.diff(xy1[0])))
    xy1[0] = ((xy1[0] - x_min)/dxy)
    xy1[1] = ((xy1[1] - y_min)/dxy)

    if dtype == int:
        xy1 = np.rint(xy1).astype(int)

    return xy1, dxy/10**digits, x_min/10**digits, y_min/10**digits


def point_xy_to_map_coords(xy, dxy, x_min, y_min, dtype=int, copy=True):
    """
    Convert an array of irregular x and y values to map_coordinates appropriate values.

    Parameters
    ----------
    xy : ndarray
        of shape (l, 2) where l is the number of location pairs. The first value of each pair is x and the second is y.
    dxy
        The grid interval.
    x_min
        The x value at index 0.
    y_min
        The y value at index 0.
    copy : bool
        Should it make a copy or modify the array in place?

    Returns
    -------
    ndarray
        Transposed shape from the input xy to an index array.
    """
    if copy:
        xy1 = xy.T.copy()
    else:
        xy1 = xy.T

    x = ((xy1[0] - x_min)/dxy)
    y = ((xy1[1] - y_min)/dxy)
    xy1[0] = y
    xy1[1] = x

    if dtype == int:
        xy1 = np.rint(xy1).astype(int)

    return xy1


def map_coords_to_xy(coords, dxy, x_min, y_min, copy=True):
    """
    The reverse of point_xy_to_map_coords.
    """
    if copy:
        coords1 = coords.copy().astype(float)
    else:
        coords1 = coords.astype(float)
    coords1[0] = coords1[0]*dxy + x_min
    coords1[1] = coords1[1]*dxy + y_min

    return coords1.T


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


#def convert_crs(from_crs, crs_type='proj4', pass_str=False):
#    """
#    Convenience function to convert one crs format to another.
#
#    Parameters
#    ----------
#    from_crs: int or str
#        The crs as either an epsg number or a str in a common crs format (e.g. proj4 or wkt).
#    crs_type: str
#        Output format type of the crs ('proj4', 'wkt', 'proj4_dict', or 'netcdf_dict').
#    pass_str: str
#        If input is a str, should it be passed though without conversion?
#
#    Returns
#    -------
#    str or dict
#    """
#
#    ### Load in crs
#    if all([pass_str, isinstance(from_crs, str)]):
#        crs2 = from_crs
#    else:
#        if isinstance(from_crs, int):
#            crs1 = parse.from_epsg_code(from_crs)
#        elif isinstance(from_crs, str):
#            crs1 = parse.from_unknown_text(from_crs)
#        else:
#            raise  ValueError('from_crs must be an int or str')
#
#        ### Convert to standard formats
#        if crs_type == 'proj4':
#            crs2 = crs1.to_proj4()
#        elif crs_type == 'wkt':
#            crs2 = crs1.to_ogc_wkt()
#        elif crs_type in ['proj4_dict', 'netcdf_dict']:
#            crs1a = crs1.to_proj4()
#            crs1b = crs1a.replace('+', '').split()[:-1]
#            crs1c = dict(i.split('=') for i in crs1b)
#            crs2 = dict((i, float(crs1c[i])) for i in crs1c)
#        else:
#            raise ValueError('Select one of "proj4", "wkt", "proj4_dict", or "netcdf_dict"')
#        if crs_type == 'netcdf_dict':
#            crs3 = {}
#            for i in crs2:
#                if i in proj4_netcdf_var.keys():
#                    t1 = proj4_netcdf_var[i]
#                    if isinstance(t1, tuple):
#                        crs3.update({j: crs2[i] for j in t1})
#                    else:
#                        crs3.update({proj4_netcdf_var[i]: crs2[i]})
#            if crs3['transform_name'] in proj4_netcdf_name.keys():
#                gmn = proj4_netcdf_name[crs3['transform_name']]
#                crs3.update({'transform_name': gmn})
#            else:
#                raise ValueError('No appropriate netcdf projection.')
#            crs2 = crs3
#
#    return crs2



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 14 09:08:58 2019

@author: MichaelEK
"""
import os
import numpy as np
import pandas as pd
import xarray as xr
from pyproj import Proj, transform
from datetime import datetime
from pdsql import mssql
from hydrointerp import Interp
from matplotlib import pyplot

pd.options.display.max_columns = 10


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]

########################################
### Parameters

dataset_type = [15]

server = 'edwprod01'
db = 'hydro'


py_dir = os.path.realpath(os.path.dirname(__file__))

nc1 = 'nasa_gpm_2017-07-20.nc'

nc_dir = r'N:\nasa\precip\gpm_3IMERGHH'
nc2 = 'gpm_3IMERGHH_v06_20150601-20150630.nc4'
nc1 = r'N:\nasa\precip\gpm_3IMERGHH\gpm_3IMERGHH_v06_20150101-20150131.nc4'

point_time_name = 'DateTime'
point_x_name = 'NZTMX'
point_y_name = 'NZTMY'
point_data_name = 'Value'
point_crs = 2193

grid_time_name = 'time'
grid_x_name = 'lon'
grid_y_name = 'lat'
grid_data_name = 'precipitationCal'
grid_crs = 4326

grid_res=20000
to_crs=2193
bbox=None
order=3
digits=2
min_val=0

########################################
### Get data

## Nasa data
ds1 = xr.open_dataset(nc1)

grid_data = ds1[[grid_data_name]].resample(time='D', closed='right', label='left').sum('time') / 2

ds1.close()
del ds1

## Met station data
summ1 = mssql.rd_sql(server, db, 'TSDataNumericDailySumm', where_in={'DatasetTypeID': dataset_type}).drop('ModDate', axis=1)
summ1.ToDate = pd.to_datetime(summ1.ToDate)
summ1.FromDate = pd.to_datetime(summ1.FromDate)

dates1 = grid_data.time.to_index()
min_date = dates1.min()
max_date = dates1.max()

summ2 = summ1[(summ1.FromDate <= min_date) & (summ1.ToDate >= max_date)].copy()

ts_data = mssql.rd_sql(server, db, 'TSDataNumericDaily', ['ExtSiteID', 'DateTime', 'Value'], where_in={'ExtSiteID': summ2.ExtSiteID.tolist(), 'DatasetTypeID': dataset_type}, from_date=str(min_date.date()), to_date=str(max_date.date()), date_col='DateTime')
ts_data.DateTime = pd.to_datetime(ts_data.DateTime)

site_data = mssql.rd_sql(server, db, 'ExternalSite', ['ExtSiteID', 'NZTMX', 'NZTMY'], where_in={'ExtSiteID': summ2.ExtSiteID.tolist()}).round()

point_data = pd.merge(ts_data, site_data, on='ExtSiteID').drop('ExtSiteID', axis=1)


self = Interp(grid_data, grid_time_name, grid_x_name, grid_y_name, grid_data_name, grid_crs, point_data, point_time_name, point_x_name, point_y_name, point_data_name, point_crs)

point_grid = self.points_to_grid(20000, to_crs)

new_grid = self.adjust_grid_from_points(20000, to_crs)


point_grid.precip.isel(time=0).plot()

new_grid.precip.isel(time=0).plot()

#############################3
### Other testing
### A 'precipitationQualityIndex' of > 0.4 is the ideal setting for GPM data

ds1 = xr.open_dataset(os.path.join(nc_dir, nc2))

ds2 = ds1[[grid_data_name, 'precipitationQualityIndex']].sel(time=slice('2015-06-18', '2015-06-19')).copy()

ds1.close()
del ds1

da1 = ds2[grid_data_name].resample(time='D', closed='right', label='left').sum('time') / 2
da2 = ds2['precipitationQualityIndex'].resample(time='D', closed='right', label='left').mean('time')

ds2b = xr.merge([da1, da2])

ds2b.to_netcdf(r'E:\ecan\git\hydrointerp\hydrointerp\datasets\nasa_gpm_2015-06-18.nc')

ds3 = da1.where(da2 > 0.4).to_dataset()

max_x = site_data.NZTMX.max()
min_x = site_data.NZTMX.min()
max_y = site_data.NZTMY.max()
min_y = site_data.NZTMY.min()

i1 = Interp(ds3, grid_time_name, grid_x_name, grid_y_name, grid_data_name, grid_crs)

i1.grid_interp_na()

ds4 = i1.grid_to_grid(10000, to_crs, bbox=(1300000, 1700000, 5000000, 5370000), min_val=0)

ds3.mean(['lat', 'lon']).round()

ds4.mean(['y', 'x']).round()

ds5 = ds4.sel(time=slice('2015-06-18', '2015-06-19'))

ds5.precip.plot(x='x', y='y', col='time', col_wrap=2)


ds3a = ds3.sel(time=slice('2015-06-18', '2015-06-19'))

ds3a[grid_data_name].plot(x='lon', y='lat', col='time', col_wrap=2)



plt1 = da5.plot(x='lon', y='lat', col='time', col_wrap=3)




grid_data = ds3.sel(time=slice('2015-06-18', '2015-06-19'))

## Met station data
summ1 = mssql.rd_sql(server, db, 'TSDataNumericDailySumm', where_in={'DatasetTypeID': dataset_type}).drop('ModDate', axis=1)
summ1.ToDate = pd.to_datetime(summ1.ToDate)
summ1.FromDate = pd.to_datetime(summ1.FromDate)

dates1 = grid_data.time.to_index()
min_date = dates1.min()
max_date = dates1.max()

summ2 = summ1[(summ1.FromDate <= min_date) & (summ1.ToDate >= max_date) & (summ1.ExtSiteID.astype(int) < 900000)].copy()

ts_data = mssql.rd_sql(server, db, 'TSDataNumericDaily', ['ExtSiteID', 'DateTime', 'Value'], where_in={'ExtSiteID': summ2.ExtSiteID.tolist(), 'DatasetTypeID': dataset_type}, from_date=str(min_date.date()), to_date=str(max_date.date()), date_col='DateTime')
ts_data.DateTime = pd.to_datetime(ts_data.DateTime)

site_data = mssql.rd_sql(server, db, 'ExternalSite', ['ExtSiteID', 'NZTMX', 'NZTMY'], where_in={'ExtSiteID': summ2.ExtSiteID.tolist()}).round()

point_data = pd.merge(ts_data, site_data, on='ExtSiteID').drop('ExtSiteID', axis=1)
# point_data.rename(columns={'Value': 'precip', 'DateTime': 'date'}, inplace=True)

point_data.to_csv(r'E:\ecan\git\hydrointerp\hydrointerp\datasets\ecan_data_2015-06-18.csv', index=False)


self = Interp(grid_data, grid_time_name, grid_x_name, grid_y_name, grid_data_name, grid_crs, point_data, point_time_name, point_x_name, point_y_name, point_data_name, point_crs)

self.grid_interp_na()

#point_grid = self.points_to_grid(10000, to_crs)

new_grid = self.adjust_grid_from_points(10000, to_crs)


new_grid.precip.plot(x='x', y='y', col='time', col_wrap=2)


comp_res = self.validate_grid_from_points(0.08, 10000, to_crs)
comp_res




























# -*- coding: utf-8 -*-
"""
Created on Mon Jan 14 09:08:58 2019

@author: MichaelEK
"""
import os
import numpy as np
import pandas as pd
import xarray as xr
from hydrolm import LM
from hilltoppy import web_service as ws
from hydrointerp.interp2d import grid_to_points, points_to_grid, grid_to_grid
from pyproj import Proj, transform
from datetime import datetime

pd.options.display.max_columns = 10


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]

########################################
### Parameters

ecan_base_url = 'http://testwateruse.ecan.govt.nz'
hts_name = 'AtmosRecorderEcanDaily.hts'

py_dir = os.path.realpath(os.path.dirname(__file__))

nc1 = 'tests/nasa_gpm_2017-07-20.nc'


nc1 = r'N:\nasa\precip\trmm_3B42\trmm_3B42_v07_20000101-20001231.nc4'
nasa_mtype = 'precipitation'
nasa_lat = 'lat'
nasa_lon = 'lon'
nasa_time = 'time'
nasa_crs = 4326
nz_crs = 2193

mtype = 'Precipitation'

from_date = '2000-01-01'
to_date = '2000-12-31'

min_n_days = 5 * 365
min_missing_days = 14

########################################
### Get data

## Nasa data
ds1 = xr.open_dataset(nc1)

da1 = ds1[nasa_mtype].resample(time='D', closed='right', label='left').sum('time') * 3

del ds1

## Met stations
sites = ws.site_list(ecan_base_url, hts_name, True)
sites.rename(columns={'SiteName': 'Site'}, inplace=True)

mtypes = []
for s in sites['Site'].values:
    mtypes.append(ws.measurement_list(ecan_base_url, hts_name, s, mtype))

mtypes_df = pd.concat(mtypes).reset_index().drop('DataType', axis=1)

mtypes_df['n_days'] = (mtypes_df['To'] - mtypes_df['From']).dt.days

mtypes_df1 = mtypes_df[mtypes_df['n_days'] >= min_n_days].copy()

sites1 = pd.merge(sites, mtypes_df1, on='Site')

ts_data = []
for index, row in sites1.iterrows():
    ts_data.append(ws.get_data(ecan_base_url, hts_name, row['Site'], row['Measurement']))

ts_data_df = pd.concat(ts_data)
ts_data_df.index = ts_data_df.index.droplevel('Measurement')

ts_data_df2 = ts_data_df['Value'].unstack(0)


#################################################
### Process station data

period_data1 = ts_data_df2[from_date:to_date].copy()
missing_days = period_data1.isnull().sum()
good_ones = missing_days[missing_days <= min_missing_days].reset_index()['Site']

good_sites1 = period_data1.loc[:, period_data1.columns.isin(good_ones)].interpolate()

period_data1.loc[:, period_data1.columns.isin(good_ones)] = good_sites1

ts_data1 = period_data1.stack().reorder_levels([1, 0]).sort_index()
ts_data1.name = 'precip'

ts_data2 = pd.merge(sites, ts_data1.reset_index(), on='Site').drop('Site', axis=1).rename(columns={'Easting': 'x', 'Northing': 'y', 'DateTime': 'time'})

##############################################
### Interpolations

points1 = sites1.rename(columns={'Easting': 'x', 'Northing': 'y'})[['x', 'y']].copy()

nasa_df1 = grid_to_points(da1.to_dataset(), nasa_time, nasa_lon, nasa_lat, nasa_mtype, points1, nasa_crs, nz_crs, order=1, min_val=0)

nasa_df2 = nasa_df1.dropna().reset_index().copy()
nasa_df2['x'] = nasa_df2['x'].round().astype(int)
nasa_df2['y'] = nasa_df2['y'].round().astype(int)

both1 = pd.merge(ts_data2, nasa_df2, on=['x', 'y', 'time'])

both1['diff'] = both1['precip'] - both1[nasa_mtype]

points2 = both1[['x', 'y']].drop_duplicates()

from_crs1 = Proj(convert_crs(nz_crs, pass_str=True), preserve_units=True)
to_crs1 = Proj(convert_crs(nasa_crs, pass_str=True), preserve_units=True)
points_wgs = np.array([transform(from_crs1, to_crs1, x, y) for x, y in points2.values]).round(3)
min_p = points_wgs.min(0)
max_p = points_wgs.max(0)

min_lon = find_nearest(da1.lon, min_p[0])
max_lon = find_nearest(da1.lon, max_p[0])
min_lat = find_nearest(da1.lat, min_p[1])
max_lat = find_nearest(da1.lat, max_p[1])

dxy = np.median(np.diff(da1[nasa_lon]))


diff_grid = points_to_grid(both1, 'time', 'x', 'y', 'diff', dxy, nz_crs, nasa_crs, bbox=(min_lon, max_lon, min_lat, max_lat))
diff_grid = diff_grid.rename({'x': nasa_lon, 'y': nasa_lat})

da2 = da1.sel({nasa_lon: da1[nasa_lon].isin(diff_grid[nasa_lon]), nasa_lat: da1[nasa_lat].isin(diff_grid[nasa_lat])})

da3 = da1 + diff_grid['diff']
da3.name = 'precipitation_corr'

ar3 = da3.values
ar3[ar3 < 0] = 0

da3.data = ar3

da3.isel(time=2).plot(x='lon', y='lat')

ds4 = grid_to_grid(da1.to_dataset(), nasa_time, nasa_lon, nasa_lat, nasa_mtype, dxy, nasa_crs, bbox=(min_lon, max_lon, min_lat, max_lat), order=1, min_val=0)

ds4[nasa_mtype].isel(time=2).plot(x='x', y='y')

site_grid = points_to_grid(both1, 'time', 'x', 'y', 'precip', dxy, nz_crs, nasa_crs, bbox=(min_lon, max_lon, min_lat, max_lat))

site_grid['precip'].isel(time=2).plot(x='x', y='y')








# -*- coding: utf-8 -*-
"""
Created on Thu Jan 10 13:08:28 2019

@author: MichaelEK
"""
import numpy as np
import pandas as pd
import xarray as xr
from hydrointerp import Interp, datasets
#from hydrointerp.io.raster import save_geotiff

#########################################
### Parameters

#py_dir = os.path.realpath(os.path.dirname(__file__))

nc1 = 'nasa_gpm_2015-06-18'
csv1 = 'ecan_data_2015-06-18'
#tif0 = 'nasa_gpm_2017-07-20.tif'


grid_time_name = 'time'
grid_x_name = 'lon'
grid_y_name = 'lat'
grid_data_name = 'precipitationCal'
grid_crs = 4326

point_time_name = 'date'
point_x_name = 'NZTMX'
point_y_name = 'NZTMY'
point_data_name = 'precip'
point_crs = 2193

to_crs = 2193
grid_res = 10000
bbox=None
order=2
extrapolation='constant'
cval=np.nan
digits = 2
min_lat = -48
max_lat = -41
min_lon = 170
max_lon = 178
min_val=0
method='linear'

########################################
### Read data

ds = xr.open_dataset(datasets.get_path(nc1))
df1 = pd.read_csv(datasets.get_path(csv1), parse_dates=['date'], infer_datetime_format=True)

### Assign nan toplaces wherethe quality index is below 0.4
ds2 = ds[[grid_data_name]].where(ds.precipitationQualityIndex > 0.4)

### Close the file (by removing the object)
ds.close()

### Convert to DataFrame
df5 = ds2[grid_data_name].to_dataframe().reset_index()

### Create example points
points_df = df1.loc[[6, 15, 132], [point_x_name, point_y_name]].copy()
points_df.rename(columns={point_x_name: 'x', point_y_name: 'y'}, inplace=True)

#def test_save_geotiff():
#    save_geotiff(df5, from_crs, 'precipitationCal', 'lon', 'lat', export_path=tif0)
#
#    assert 0 == 0

########################################
### Run interpolations

interpc = Interp(ds2, grid_time_name, grid_x_name, grid_y_name, grid_data_name, grid_crs, point_data=df1, point_time_name=point_time_name, point_x_name=point_x_name, point_y_name=point_y_name, point_data_name=point_data_name, point_crs=point_crs)
interpc.grid_interp_na()


def test_grid_interp_na():
    nan1 = ds2[grid_data_name].isnull().sum()
    nan2 = interpc.grid_data['precip'].isnull().sum()
    assert (nan1 > 0) & (nan2 == 0)


def test_grid_to_grid2():
    interp1 = interpc.grid_to_grid(grid_res, to_crs, bbox, order, extrapolation, min_val=min_val)
    assert 28 > interp1.precip.mean() > 27


def test_points_to_grid2():
    interp2 = interpc.points_to_grid(grid_res, to_crs, bbox, method, extrapolation, min_val=min_val)
    assert 28 > interp2.precip.mean() > 27


def test_grid_to_points2():
    interp3 = interpc.grid_to_points(points_df, to_crs, order, min_val=min_val)
    assert 149 > interp3.precip.mean() > 148


def test_points_to_points2():
    interp4 = interpc.points_to_points(points_df, to_crs, method, min_val=min_val)
    assert 34 > interp4.precip.mean() > 33


def test_adj_grid_from_points():
    interp5 = interpc.adjust_grid_from_points(grid_res, to_crs)
    assert 55 > interp5.precip.mean() > 54


def test_validate_grid_from_points():
    interp6 = interpc.validate_grid_from_points(0.08, grid_res, to_crs)
    assert len(interp6) > 8




# -*- coding: utf-8 -*-
"""
Raster and spatial interpolation functions.
"""
import pandas as pd
import numpy as np
try:
    import fiona
    _fiona = True
except:
    _fiona = False
import xarray as xr
from scipy.interpolate import griddata, RectBivariateSpline
from pyproj import Proj, CRS, Transformer
from scipy.ndimage import map_coordinates
from hydrointerp.util import grid_xy_to_map_coords, point_xy_to_map_coords
#from util import grid_xy_to_map_coords, point_xy_to_map_coords

#################################################
### Helper functions


def _process_grid_input(input_data, time_name, x_name, y_name, data_name, zero_pad=True):
    if isinstance(input_data, pd.DataFrame):
        grouped = input_data.set_index([time_name, y_name, x_name])[data_name].sort_index()
        # grouped = input_data.set_index([time_name, x_name, y_name])[data_name].sort_index()
        input_data = grouped.to_xarray().to_dataset()
    if isinstance(input_data, xr.Dataset):
        da1 = input_data[data_name]
        da2 = da1.transpose(time_name, y_name, x_name).sortby([time_name, y_name, x_name])
        # da2 = da1.transpose(time_name, x_name, y_name).sortby([time_name, x_name, y_name])
        arr1 = da2.values
        if zero_pad:
            np.nan_to_num(arr1, False)
        x = da1[x_name].values
        y = da1[y_name].values
        time1 = da1[time_name].values
        # xy_orig_pts = np.dstack(np.meshgrid(y, x)).reshape(-1, 2)
        xy_orig_pts = np.dstack(np.meshgrid(x, y)).reshape(-1, 2)

    return arr1, xy_orig_pts, time1

################################################
### Core functions


def grid_to_grid(grid, time_name, x_name, y_name, data_name, grid_res, from_crs, to_crs=None, bbox=None, order=3, extrapolation='constant', fill_val=np.nan, digits=2, min_val=None):
    """
    Function to interpolate regularly or irregularly spaced values over many time stamps. Each time stamp of spatial values are interpolated independently (2D interpolation as opposed to 3D interpolation). Returns an xarray Dataset with the 3 dimensions. Uses the scipy interpolation function called map_coordinates.

    Parameters
    ----------
    grid : DataFrame or Dataset
        A pandas DataFrame or an xarray Dataset. It's recommended to use an xarray Dataset for the input grid as it ensures that the user knows that it is truly regular. Regardless, the input will be regularised.
    time_name : str
        If grid is a DataFrame, then time_name is the time column name. If grid is a Dataset, then time_name is the time coordinate name.
    x_name : str
        If grid is a DataFrame, then x_name is the x column name. If grid is a Dataset, then x_name is the x coordinate name.
    y_name : str
        If grid is a DataFrame, then y_name is the y column name. If grid is a Dataset, then y_name is the y coordinate name.
    data_name : str
        If grid is a DataFrame, then data_name is the data column name. If grid is a Dataset, then data_name is the data variable name.
    grid_res : int or float
        The resulting grid resolution in the unit of the final projection (usually meters or decimal degrees).
    from_crs : int or str or None
        The projection info for the input data if the result should be reprojected to the to_crs projection (either a proj4 str or epsg int).
    to_crs : int or str or None
        The projection for the output data similar to from_crs.
    bbox : tuple of int or float
        The bounding box for the output interpolation in the to_crs projection. None will return a similar grid extent as the input. The tuple should contain four ints or floats in the following order: (x_min, x_max, y_min, y_max)
    order : int
        The order of the spline interpolation, default is 3. The order has to be in the range 0-5. An order of 1 is linear interpolation.
    extrapolation : str
        The equivalent of 'mode' in the map_coordinates function. Options are: 'constant', 'nearest', 'reflect', 'mirror', and 'wrap'. Most reseaonable options for this function will be either 'constant' or 'nearest'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.map_coordinates.html>`_ for more details.
    fill_val : int or float
        If 'constant' if passed to the extrapolation parameter, fill_val assigns the value outside of the boundary. Defaults to numpy.nan.
    digits : int
        The number of digits to round the output.
    min_val : int, float, or None
        The minimum value for the results. All results below min_val will be assigned min_val.

    Returns
    -------
    xarray Dataset
    """
    print('Preparing input and output')
    if from_crs == 4326:
        input_digits = 4
    else:
        input_digits = 0
    if (to_crs == 4326) | ((from_crs == 4326) & (isinstance(to_crs, (str, int, dict)))):
        output_digits = 4
    else:
        output_digits = 0

    ### Prepare input data
    arr1, xy_orig_pts, time1 = _process_grid_input(grid, time_name, x_name, y_name, data_name)

    input_coords, dxy, x_min, y_min = grid_xy_to_map_coords(xy_orig_pts, input_digits)

    ### convert to new projection and prepare X/Y data
    if isinstance(from_crs, (str, int, dict)) & isinstance(to_crs, (str, int, dict)):
        from_crs1 = Proj(CRS.from_user_input(from_crs))
        to_crs1 = Proj(CRS.from_user_input(to_crs))
        trans1 = Transformer.from_proj(from_crs1, to_crs1, always_xy=True)
        xy_new = np.array(trans1.transform(*xy_orig_pts.T)).round(output_digits)
        # out_y_min, out_x_min = xy_new.min(1)
        # out_y_max, out_x_max = xy_new.max(1)
        out_x_min, out_y_min = xy_new.min(1)
        out_x_max, out_y_max = xy_new.max(1)
    else:
        out_x_max, out_y_max = xy_orig_pts.max(0)
        out_x_max, out_y_max = xy_orig_pts.max(0)
        out_x_min = x_min
        out_y_min = y_min

    ### Prepare output data
    if isinstance(bbox, tuple):
        out_x_min, out_x_max, out_y_min, out_y_max = bbox

    new_x = np.arange(out_x_min, out_x_max, grid_res)
    new_y = np.arange(out_y_min, out_y_max, grid_res)

    # xy_out = np.dstack(np.meshgrid(new_y, new_x)).reshape(-1, 2)
    xy_out = np.dstack(np.meshgrid(new_x, new_y)).reshape(-1, 2)

    if isinstance(to_crs, (str, int, dict)):
        trans2 = Transformer.from_proj(to_crs1, from_crs1, always_xy=True)
        xy_new_index = np.array(trans2.transform(*xy_out.T)).T
    else:
        xy_new_index = xy_out

    output_coords = point_xy_to_map_coords(xy_new_index, dxy, x_min, y_min, float)

    ### Run interpolation (need to add mutliprossessing)
    arr2 = np.zeros((len(time1), output_coords.shape[1]), arr1.dtype)

    print('Running interpolations...')
    ## An example for using RectBivariateSpline as the equivalent to map_coordinates output (about half as fast):
#    arr2a = arr2.copy()
#
#    x_out = xy_new_index.T[1]
#    y_out = xy_new_index.T[0]
#
#    for d in np.arange(len(arr1)):
#        arr2a[d] = RectBivariateSpline(y, x, arr1[d], kx=3, ky=3).ev(y_out, x_out)

    for d in np.arange(len(arr1)):
         map_coordinates(arr1[d], output_coords, arr2[d], order=order, mode=extrapolation, cval=fill_val, prefilter=True)

    ### Reshape and package data
    print('Packaging up the output')
    arr2 = arr2.reshape((len(time1), len(new_y), len(new_x))).round(digits)

    if isinstance(min_val, (int, float)):
        arr2[arr2 < min_val] = min_val

    new_ds = xr.DataArray(arr2, coords=[time1, new_y, new_x], dims=['time', 'y', 'x'], name=data_name).to_dataset()

    return new_ds


def grid_to_points(grid, time_name, x_name, y_name, data_name, point_data, from_crs, to_crs=None, order=3, digits=2, min_val=None):
    """
    Function to take a dataframe of point value inputs (df) and interpolate to other points (point_data). Uses the scipy interpolation function called map_coordinate.

    Parameters
    ----------
    grid : DataFrame or Dataset
        A pandas DataFrame or an xarray Dataset. It's recommended to use an xarray Dataset for the input grid as it ensures that the user knows that it is truly regular. Regardless, the input will be regularised.
    time_name : str
        If grid is a DataFrame, then time_name is the time column name. If grid is a Dataset, then time_name is the time coordinate name.
    x_name : str
        If grid is a DataFrame, then x_name is the x column name. If grid is a Dataset, then x_name is the x coordinate name.
    y_name : str
        If grid is a DataFrame, then y_name is the y column name. If grid is a Dataset, then y_name is the y coordinate name.
    data_name : str
        If grid is a DataFrame, then data_name is the data column name. If grid is a Dataset, then data_name is the data variable name.
    point_data : str or DataFrame
        Path to geometry file of points to be interpolated (e.g. shapefile). Can be any file type that fiona/gdal can open. It can also be a DataFrame with 'x' and 'y' columns and the crs must be the same as to_crs.
    from_crs : int or str or None
        The projection info for the input data if the result should be reprojected to the to_crs projection (either a proj4 str or epsg int).
    to_crs : int or str or None
        The projection for the output data similar to from_crs.
    order : int
        The order of the spline interpolation, default is 3. The order has to be in the range 0-5. An order of 1 is linear interpolation.
    digits : int
        The number of digits to round the output.
    min_val : int, float, or None
        The minimum value for the results. All results below min_val will be assigned min_val.

    Returns
    -------
    DataFrame
    """
    print('Preparing input and output')
    if from_crs == 4326:
        input_digits = 4
    else:
        input_digits = 0

    ### Read in points
    if isinstance(point_data, str):
        if _fiona:
            with fiona.open(point_data) as f1:
                point_crs = Proj(f1.crs)
                points = np.array([tuple(p['geometry']['coordinates']) for p in f1 if p['geometry']['type'] == 'Point'])
        else:
            raise ImportError('Please install fiona for importing GIS files')
    elif isinstance(point_data, pd.DataFrame):
        point_crs = CRS.from_user_input(to_crs)
        points = point_data[['x', 'y']].values
    else:
        raise ValueError('point_path must be a str path to a geometry file (e.g. shapefile) or a DataFrame with the same x_name and y_name columns')

    ### Prepare input grid
    arr1, xy_orig_pts, time1 = _process_grid_input(grid, time_name, x_name, y_name, data_name)

    input_coords, dxy, x_min, y_min = grid_xy_to_map_coords(xy_orig_pts, input_digits)

    ### Convert points to to_crs, from_crs, and array index
    from_crs1 = CRS.from_user_input(from_crs)
    if to_crs is not None:
        to_crs1 = CRS.from_user_input(to_crs)
        trans1 = Transformer.from_crs(point_crs, to_crs1)
        points = np.array(trans1.transform(*points.T)).T
    else:
        to_crs1 = point_crs
    trans2 = Transformer.from_crs(to_crs1, from_crs1)
    points_from_crs = np.array(trans2.transform(*points.T)).T
    points_coords = point_xy_to_map_coords(points_from_crs, dxy, x_min, y_min, float)

    ### Run interpolations
    print('Running interpolations...')
    arr2 = np.zeros((len(time1), points_coords.shape[1]), arr1.dtype)

    for d in np.arange(len(arr1)):
         map_coordinates(arr1[d], points_coords, arr2[d], order=order, cval=np.nan, prefilter=True)

    if isinstance(min_val, (int, float)):
        arr2[arr2 < min_val] = min_val

    ### Reshape and package data
    print('Packaging up the output')
    arr3 = arr2.flatten().round(digits)

    time_ar = np.repeat(time1, len(points))
    y_ar = np.tile(points.T[1], len(time1))
    x_ar = np.tile(points.T[0], len(time1))
    new_df = pd.DataFrame({'time': time_ar, 'y': y_ar, 'x': x_ar, data_name: arr3}).set_index(['time', 'y', 'x'])

    return new_df


def points_to_grid(df, time_name, x_name, y_name, data_name, grid_res, from_crs, to_crs=None, bbox=None, method='linear', extrapolation='contstant', fill_val=np.nan, digits=2, min_val=None):
    """
    Function to take a dataframe of point value inputs (df) and interpolate to a grid. Uses the scipy griddata function for interpolation.

    Parameters
    ----------
    df : DataFrame
        A pandas DataFrame containing four columns as shown in the below parameters.
    time_name : str
        If grid is a DataFrame, then time_name is the time column name. If grid is a Dataset, then time_name is the time coordinate name.
    x_name : str
        If grid is a DataFrame, then x_name is the x column name. If grid is a Dataset, then x_name is the x coordinate name.
    y_name : str
        If grid is a DataFrame, then y_name is the y column name. If grid is a Dataset, then y_name is the y coordinate name.
    data_name : str
        If grid is a DataFrame, then data_name is the data column name. If grid is a Dataset, then data_name is the data variable name.
    grid_res : int or float
        The resulting grid resolution in the unit of the final projection (usually meters or decimal degrees).
    from_crs : int or str or None
        The projection info for the input data if the result should be reprojected to the to_crs projection (either a proj4 str or epsg int).
    to_crs : int or str or None
        The projection for the output data similar to from_crs.
    bbox : tuple of int or float
        The bounding box for the output interpolation in the to_crs projection. None will return a similar grid extent as the input. The tuple should contain four ints or floats in the following order: (x_min, x_max, y_min, y_max)
    method : str
        The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
    extrapolation : str
        Either 'constant' or 'nearest'.
    fill_val : int or float
        If 'constant' if passed to the extrapolation parameter, fill_val assigns the value outside of the boundary. Defaults to numpy.nan.
    digits : int
        The number of digits to round the output.
    min_val : int, float, or None
        The minimum value for the results. All results below min_val will be assigned min_val.

    Returns
    -------
    xarray Dataset
    """
    print('Prepare input and output data')
    if (to_crs == 4326) | ((from_crs == 4326) & (to_crs is None)):
        output_digits = 4
    else:
        output_digits = 0

    ### Prepare input data
    df2 = df.copy()

    time1 = pd.to_datetime(df2[time_name].sort_values().unique())

    ### Convert input data to crs of points shp and create input xy
    from_crs1 = Proj(CRS.from_user_input(from_crs))
    xy1 = np.array(list(zip(df2[x_name], df2[y_name]))).T
    if isinstance(to_crs, (str, int, dict)):
        to_crs1 = Proj(CRS.from_user_input(to_crs))
        trans1 = Transformer.from_proj(from_crs1, to_crs1)
        xy1 = np.array(trans1.transform(*xy1))
        df2[x_name] = xy1[0]
        df2[y_name] = xy1[1]

    ### Prepare output data
    if isinstance(bbox, tuple):
        out_x_min, out_x_max, out_y_min, out_y_max = bbox
    else:
        out_x_min, out_y_min = xy1.min(1).round(output_digits)
        out_x_max, out_y_max = xy1.max(1).round(output_digits)

    new_x = np.arange(out_x_min, out_x_max, grid_res)
    new_y = np.arange(out_y_min, out_y_max, grid_res)

    xy_out = np.dstack(np.meshgrid(new_x, new_y)).reshape(-1, 2)

    ### Run interpolations
    print('Run interpolations...')
    arr2 = np.zeros((len(time1), xy_out.shape[0]), df2[data_name].dtype)
    index1 = {time1[i]: i for i in np.arange(len(time1))}

#    start1 = time()
    for name, group in df2.groupby(time_name):
        print(name)
        i = index1[name]
        xy = group[[x_name, y_name]].values
        arr2[i] = griddata(xy, group[data_name].values, xy_out, method=method, fill_value=fill_val).round(digits)
        if extrapolation == 'nearest':
            nan_index = np.isnan(arr2[i])
            nan_xy = xy_out[nan_index]
            nonnan_values = arr2[i][~nan_index]
            nonnan_xy = xy_out[~nan_index]
            arr2[i][nan_index] = griddata(nonnan_xy, nonnan_values, nan_xy, method='nearest').round(digits)
#    end1 = time()
#    setb = end1 - start1

    ### Reshape and package data
    print('Packaging up the output')
    arr2 = arr2.reshape((len(time1), len(new_y), len(new_x))).round(digits)

    if isinstance(min_val, (int, float)):
        arr2[arr2 < min_val] = min_val

    new_ds = xr.DataArray(arr2, coords=[time1, new_y, new_x], dims=['time', 'y', 'x'], name=data_name).to_dataset()

    return new_ds


def points_to_points(df, time_name, x_name, y_name, data_name, point_data, from_crs, to_crs=None, method='linear', digits=2, min_val=None):
    """
    Function to take a dataframe of point value inputs (df) and interpolate to other points (point_data). Uses the scipy griddata function for interpolation.

    Parameters
    ----------
    df : DataFrame
        A pandas DataFrame containing four columns as shown in the below parameters.
    time_name : str
        If grid is a DataFrame, then time_name is the time column name. If grid is a Dataset, then time_name is the time coordinate name.
    x_name : str
        If grid is a DataFrame, then x_name is the x column name. If grid is a Dataset, then x_name is the x coordinate name.
    y_name : str
        If grid is a DataFrame, then y_name is the y column name. If grid is a Dataset, then y_name is the y coordinate name.
    data_name : str
        If grid is a DataFrame, then data_name is the data column name. If grid is a Dataset, then data_name is the data variable name.
    point_data : str or DataFrame
        Path to geometry file of points to be interpolated (e.g. shapefile). Can be any file type that fiona/gdal can open. It can also be a DataFrame with 'x' and 'y' columns and the crs must be the same as to_crs.
    from_crs : int or str or None
        The projection info for the input data if the result should be reprojected to the to_crs projection (either a proj4 str or epsg int).
    to_crs : int or str or None
        The projection for the output data similar to from_crs.
    method : str
        The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
    fill_val : int or float
        fill_val assigns the value outside of the boundary. Defaults to numpy.nan.
    digits : int
        The number of digits to round the output.
    min_val : int, float, or None
        The minimum value for the results. All results below min_val will be assigned min_val.

    Returns
    -------
    DataFrame
    """
    ### Read in points
    if isinstance(point_data, str):
        if _fiona:
            with fiona.open(point_data) as f1:
                point_crs = Proj(f1.crs)
                points = np.array([p['geometry']['coordinates'] for p in f1 if p['geometry']['type'] == 'Point'])
        else:
            raise ImportError('Please install fiona for importing GIS files')
    elif isinstance(point_data, pd.DataFrame):
        point_crs = Proj(CRS.from_user_input(to_crs))
        points = point_data[['x', 'y']].values
    else:
        raise ValueError('point_path must be a str path to a geometry file (e.g. shapefile) or a DataFrame with the same x_name and y_name columns')

    df2 = df.copy()

    time1 = pd.to_datetime(df2[time_name].sort_values().unique())

    ### Convert input data to crs of points shp and create input xy
    if to_crs is not None:
        to_crs1 = Proj(CRS.from_user_input(to_crs))
        trans1 = Transformer.from_proj(point_crs, to_crs1)
        points = np.array(trans1.transform(*points.T))
    else:
        to_crs1 = point_crs
    from_crs1 = Proj(CRS.from_user_input(from_crs))
    xy1 = df2[[x_name, y_name]].values
    trans2 = Transformer.from_proj(from_crs1, to_crs1)
    xy_new1 = np.array(trans2.transform(*xy1.T))
    df2[x_name] = xy_new1[0]
    df2[y_name] = xy_new1[1]

    ### Run interpolations
    points1 = np.array((points[1], points[0])).T
    new_lst = []
    for name, group in df2.groupby(time_name):
        print(name)
        xy = group[[y_name, x_name]].values
        new_z = griddata(xy, group[data_name].values, points1, method=method, fill_value=np.nan).round(digits)
        if isinstance(min_val, (int, float)):
            new_z[new_z < min_val] = min_val
        new_lst.extend(new_z.tolist())

    ### Create new df
    time_ar = np.repeat(time1, len(points1))
    y_ar = np.tile(points1.T[1], len(time1))
    x_ar = np.tile(points1.T[0], len(time1))
    new_df = pd.DataFrame({'time': time_ar, 'y': y_ar, 'x': x_ar, data_name: new_lst}).set_index(['time', 'y', 'x'])

    ### Export results
    return new_df


def grid_interp_na(grid, time_name, x_name, y_name, data_name, method='linear', min_val=None):
    """
    Function to fill in nan in a grid to make it complete.

    Parameters
    ----------
    grid : DataFrame or Dataset
        A pandas DataFrame or an xarray Dataset. It's recommended to use an xarray Dataset for the input grid as it ensures that the user knows that it is truly regular. Regardless, the input will be regularised.
    time_name : str
        If grid is a DataFrame, then time_name is the time column name. If grid is a Dataset, then time_name is the time coordinate name.
    x_name : str
        If grid is a DataFrame, then x_name is the x column name. If grid is a Dataset, then x_name is the x coordinate name.
    y_name : str
        If grid is a DataFrame, then y_name is the y column name. If grid is a Dataset, then y_name is the y coordinate name.
    data_name : str
        If grid is a DataFrame, then data_name is the data column name. If grid is a Dataset, then data_name is the data variable name.
    method : str
        The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
    min_val : int, float, or None
        The minimum value for the results. All results below min_val will be assigned min_val.

    Returns
    -------
    Xarray Dataset
    """
    ### Prepare input data
    arr1, xy_orig_pts, time1 = _process_grid_input(grid, time_name, x_name, y_name, data_name, False)
    arr2 = arr1.reshape(len(time1), xy_orig_pts.shape[0])

    ### Run through interpolations
    for t in arr2:
        isnan1 = np.isnan(t)
        t[isnan1] = griddata(xy_orig_pts[~isnan1], t[~isnan1], xy_orig_pts[isnan1], method=method, fill_value=np.nan)

    ### Reshape and package data
    arr2 = arr2.reshape(arr1.shape)

    if isinstance(min_val, (int, float)):
        arr2[arr2 < min_val] = min_val

    new_ds = xr.DataArray(arr2, coords=[time1, grid[y_name].values, grid[x_name].values], dims=['time', y_name, x_name], name=data_name).to_dataset()

    return new_ds



# -*- coding: utf-8 -*-
"""
Created on Sat Apr  6 20:43:35 2019

@author: michaelek
"""
import numpy as np
import pandas as pd
import xarray as xr
from pyproj import Proj, CRS, Transformer
from hydrointerp import interp2d
#import interp2d
from hydrointerp import util
# import util
import random

###################################
### base class


class Interp(object):
    """
    Base Interp class to prepare the input data for the interpolation functions.

    Parameters
    ----------
    data : DataFrame or Dataset
        A pandas DataFrame containing four columns as shown in the below parameters or an Xarray Dataset.
    time_name : str
        If grid is a DataFrame, then time_name is the time column name. If grid is a Dataset, then time_name is the time coordinate name.
    x_name : str
        If grid is a DataFrame, then x_name is the x column name. If grid is a Dataset, then x_name is the x coordinate name.
    y_name : str
        If grid is a DataFrame, then y_name is the y column name. If grid is a Dataset, then y_name is the y coordinate name.
    data_name : str
        If grid is a DataFrame, then data_name is the data column name. If grid is a Dataset, then data_name is the data variable name.
    from_crs : int or str or None
        The projection info for the input data if the result should be reprojected to the to_crs projection (either a proj4 str or epsg int).

        Returns
        -------
        Interp object
        """
    def __init__(self, grid_data=None, grid_time_name=None, grid_x_name=None, grid_y_name=None, grid_data_name=None, grid_crs=None, point_data=None, point_time_name=None, point_x_name=None, point_y_name=None, point_data_name=None, point_crs=None):

        ## Assign variables
        self._grid_crs = grid_crs
        self._point_crs = point_crs

        ## Check data type
        if point_data is not None:
            if isinstance(point_data, pd.DataFrame):
                self.point_data = point_data.rename(columns={point_x_name: 'x', point_y_name: 'y', point_time_name: 'time', point_data_name: 'precip'}).copy()
                if self.point_data.dtypes['time'].name != 'datetime64[ns]':
                    self.point_data['time'] = pd.to_datetime(self.point_data['time'])
                self.points_to_grid = self._points_to_grid
                self.points_to_points = self._points_to_points
            else:
                raise TypeError('point_data must be a DataFrame')
        if grid_data is not None:
            if isinstance(grid_data, xr.Dataset):
                self.grid_data = grid_data.rename({grid_x_name: 'x', grid_y_name: 'y', grid_time_name: 'time', grid_data_name: 'precip'}).copy()
                if self.grid_data.coords['time'].dtype.name != 'datetime64[ns]':
                    self.grid_data['time'] = pd.to_datetime(self.grid_data['time'].to_index())
                self.grid_interp_na = self._grid_interp_na
                self.grid_to_grid = self._grid_to_grid
                self.grid_to_points = self._grid_to_points
            else:
                raise TypeError('grid_data must be a Dataset')
        if (point_data is not None) and (grid_data is not None):
            self.adjust_grid_from_points = self._adjust_grid_from_points
            self.validate_grid_from_points = self._validate_grid_from_points
        if (point_data is None) and (grid_data is None):
            raise ValueError('point_data must be a DataFrame and/or grid_data must be a Dataset')
        pass


    def _grid_to_grid(self, grid_res, to_crs=None, bbox=None, order=3, extrapolation='constant', fill_val=np.nan, digits=2, min_val=None):
        """
        Function to interpolate regularly or irregularly spaced values over many time stamps. Each time stamp of spatial values are interpolated independently (2D interpolation as opposed to 3D interpolation). Returns an xarray Dataset with the 3 dimensions. Uses the scipy interpolation function called map_coordinates.

        Parameters
        ----------
        grid_res : int or float
            The resulting grid resolution in the unit of the final projection (usually meters or decimal degrees).
        to_crs : int or str or None
            The projection for the output data similar to from_crs.
        bbox : tuple of int or float
            The bounding box for the output interpolation in the to_crs projection. None will return a similar grid extent as the input. The tuple should contain four ints or floats in the following order: (x_min, x_max, y_min, y_max)
        order : int
            The order of the spline interpolation, default is 3. The order has to be in the range 0-5. An order of 1 is linear interpolation.
        extrapolation : str
            The equivalent of 'mode' in the map_coordinates function. Options are: 'constant', 'nearest', 'reflect', 'mirror', and 'wrap'. Most reseaonable options for this function will be either 'constant' or 'nearest'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.map_coordinates.html>`_ for more details.
        fill_val : int or float
            If 'constant' if passed to the extrapolation parameter, fill_val assigns the value outside of the boundary. Defaults to numpy.nan.
        digits : int
            The number of digits to round the output.
        min_val : int, float, or None
            The minimum value for the results. All results below min_val will be assigned min_val.

        Returns
        -------
        xarray Dataset
        """
        out1 = interp2d.grid_to_grid(self.grid_data, 'time', 'x', 'y', 'precip', grid_res, self._grid_crs, to_crs, bbox, order, extrapolation, fill_val, digits, min_val)
        return out1

    def _grid_to_points(self, point_data, to_crs=None, bbox=None, order=3, digits=2, min_val=None):
        """
        Function to take a dataframe of point value inputs (df) and interpolate to other points (point_data). Uses the `scipy griddata function <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for interpolation.

        Parameters
        ----------
        point_data : str or DataFrame
            Path to geometry file of points to be interpolated (e.g. shapefile). Can be any file type that fiona/gdal can open. It can also be a DataFrame with 'x' and 'y' columns and the crs must be the same as to_crs.
        to_crs : int or str or None
            The projection for the output data similar to from_crs.
        bbox : tuple of int or float
            The bounding box for the output interpolation in the to_crs projection. None will return a similar grid extent as the input. The tuple should contain four ints or floats in the following order: (x_min, x_max, y_min, y_max)
        order : int
            The order of the spline interpolation, default is 3. The order has to be in the range 0-5. An order of 1 is linear interpolation.
        digits : int
            The number of digits to round the output.
        min_val : int, float, or None
            The minimum value for the results. All results below min_val will be assigned min_val.

        Returns
        -------
        DataFrame
        """
        out1 = interp2d.grid_to_points(self.grid_data, 'time', 'x', 'y', 'precip', point_data, self._grid_crs, to_crs, order, digits, min_val)
        return out1

    def _points_to_grid(self, grid_res, to_crs=None, bbox=None, method='linear', extrapolation='contstant', fill_val=np.nan, digits=2, min_val=None):
        """
        Function to take a dataframe of point value inputs (df) and interpolate to a grid. Uses the `scipy griddata function <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for interpolation.

    Parameters
    ----------
    grid_res : int or float
        The resulting grid resolution in the unit of the final projection (usually meters or decimal degrees).
    to_crs : int or str or None
        The projection for the output data similar to from_crs.
    bbox : tuple of int or float
        The bounding box for the output interpolation in the to_crs projection. None will return a similar grid extent as the input. The tuple should contain four ints or floats in the following order: (x_min, x_max, y_min, y_max)
    method : str
        The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
    extrapolation : str
        Either 'constant' or 'nearest'.
    fill_val : int or float
        If 'constant' if passed to the extrapolation parameter, fill_val assigns the value outside of the boundary. Defaults to numpy.nan.
    digits : int
        The number of digits to round the output.
    min_val : int, float, or None
        The minimum value for the results. All results below min_val will be assigned min_val.

        Returns
        -------
        xarray Dataset
        """
        out1 = interp2d.points_to_grid(self.point_data, 'time', 'x', 'y', 'precip', grid_res, self._point_crs, to_crs, bbox, method, extrapolation, fill_val, digits, min_val)
        return out1

    def _points_to_points(self, point_data, to_crs=None, method='linear', digits=2, min_val=None):
        """
        Function to interpolate regularly or irregularly spaced values over many time stamps. Each time stamp of spatial values are interpolated independently (2D interpolation as opposed to 3D interpolation). Returns an xarray Dataset with the 3 dimensions. Uses the scipy interpolation function called map_coordinates.

        Parameters
        ----------
        point_data : str or DataFrame
            Path to geometry file of points to be interpolated (e.g. shapefile). Can be any file type that fiona/gdal can open. It can also be a DataFrame with 'x' and 'y' columns and the crs must be the same as to_crs.
        to_crs : int or str or None
            The projection for the output data similar to from_crs.
        method : str
            The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'.
        digits : int
            The number of digits to round the output.
        min_val : int, float, or None
            The minimum value for the results. All results below min_val will be assigned min_val.

        Returns
        -------
        DataFrame
        """
        out1 = interp2d.points_to_points(self.point_data, 'time', 'x', 'y', 'precip', point_data, self._point_crs, to_crs, method, digits, min_val)
        return out1


    def _grid_interp_na(self, method='linear', min_val=None, inplace=True):
        """
        Function to fill nan's in the grid_data to make it complete. Necessary for grid_to_* functions.

        Parameters
        ----------
        method : str
            The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
        min_val : int, float, or None
            The minimum value for the results. All results below min_val will be assigned min_val.
        inplace : bool
            Should the new grid overwrite the grid_data from initialisation?

        Returns
        -------
        Xarray Dataset
        """
        new_grid = interp2d.grid_interp_na(self.grid_data, 'time', 'x', 'y', 'precip', method=method, min_val=min_val)

        if inplace:
            self.grid_data = new_grid
        else:
            return new_grid


    def _adjust_grid_from_points(self, grid_res=None, to_crs=None, order=2, method='linear', digits=2, min_val=0):
        """
        Method to adjust a grid by forcing it through point data.

        Parameters
        ----------
        grid_res : int, float, or None
            The resulting grid resolution in the unit of the final projection (usually meters or decimal degrees).
        to_crs : int or str or None
            The projection for the output data similar to from_crs.
        order : int
            The order of the spline interpolation, default is 3. The order has to be in the range 0-5. An order of 1 is linear interpolation.
        method : str
            The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
        digits : int
            The number of digits to round the output.
        min_val : int, float, or None
            The minimum value for the results. All results below min_val will be assigned min_val.

        Returns
        -------
        Dataset
        """
        ## Resample the grid if needed
        if isinstance(grid_res, (int, float)):
            grid1 = self.grid_to_grid(grid_res, to_crs=to_crs, order=order, digits=digits, min_val=min_val)
        else:
            grid1 = self.grid_data

        ## grid to points
        point_data = self.point_data.copy()
        sites1 = point_data[['x', 'y']].drop_duplicates().copy()

        pts1 = self.grid_to_points(sites1, to_crs=to_crs, order=order, digits=digits, min_val=min_val)

        ## Subtract the original points from the new points
        pts2 = pts1.dropna().reset_index().copy()
        pts2['x'] = pts2['x'].round().astype(int)
        pts2['y'] = pts2['y'].round().astype(int)

        # Convert crs if needed
        if (to_crs is not None) & (to_crs != self._point_crs):
            to_crs1 = Proj(CRS.from_user_input(to_crs))
            trans1 = Transformer.from_proj(self._point_crs, to_crs1)
            points = np.array(trans1.transform(*point_data[['x', 'y']].values.T))
            point_data['x'] = points[0]
            point_data['y'] = points[1]

        both1 = pd.merge(point_data, pts2.rename(columns={'precip': 'grid_precip'}), on=['x', 'y', 'time'])

        both1['ratio'] = both1['precip']/both1['grid_precip']
        both1.loc[both1['ratio'].isnull(), 'ratio'] = 0
        self.ratio_precip = both1

        ## Create the bounding box for the new grid
        min_p = sites1.min(0)
        max_p = sites1.max(0)

        min_lon = util.find_nearest(grid1.x, min_p[0])
        max_lon = util.find_nearest(grid1.x, max_p[0])
        min_lat = util.find_nearest(grid1.y, min_p[1])
        max_lat = util.find_nearest(grid1.y, max_p[1])

        ## Points to grid
        ratio_grid = interp2d.points_to_grid(both1[['time', 'x', 'y', 'ratio']], 'time', 'x', 'y', 'ratio', grid_res, to_crs, None, (min_lon, max_lon, min_lat, max_lat), method, 'nearest')
        val = ratio_grid.ratio.values
        val[val == np.inf] = 0

        ## Add original grid by diff grid
        grid2 = xr.merge([grid1, ratio_grid], join='inner')
        grid3 = grid2['precip'] * grid2['ratio']
        grid3.name = 'precip'
        if isinstance(min_val, (int, float)):
            grid3 = xr.where(grid3 < min_val, 0, grid3).copy()

        ## Return
        return grid3.to_dataset()


    def _validate_grid_from_points(self, test_perc=0.1, grid_res=None, to_crs=None, order=2, method='linear', digits=2, min_val=0):
        """
        Method to validate a grid created by the adjust_grid_from_points method.

        Parameters
        ----------
        grid_res : int, float, or None
            The resulting grid resolution in the unit of the final projection (usually meters or decimal degrees).
        to_crs : int or str or None
            The projection for the output data similar to from_crs.
        order : int
            The order of the spline interpolation, default is 3. The order has to be in the range 0-5. An order of 1 is linear interpolation.
        method : str
            The scipy griddata interpolation method to be applied. Options are 'nearest', 'linear', and 'cubic'. See `<https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.griddata.html>`_ for more details.
        digits : int
            The number of digits to round the output.
        min_val : int, float, or None
            The minimum value for the results. All results below min_val will be assigned min_val.

        Returns
        -------
        DataFrame
        """
        ## Prepare input data
        point_data = self.point_data.copy()
        sites1 = point_data[['x', 'y']].drop_duplicates().reset_index(drop=True).copy()
        rem_sites = random.sample(range(len(sites1)), int(len(sites1)*test_perc))
        new_sites1 = sites1.loc[~sites1.index.isin(rem_sites)]
        new_point_data = pd.merge(point_data, new_sites1, on=['x', 'y'])

        ## Create grid from grid and points
        self.point_data = new_point_data
        grid_grid = self.adjust_grid_from_points(grid_res, to_crs, order, method, digits, min_val)
        self.point_data = point_data

        ## Create points from new points
        rem_sites1 = sites1.loc[sites1.index.isin(rem_sites)]
        rem_point_data = pd.merge(point_data, rem_sites1, on=['x', 'y'])

        points_points = interp2d.points_to_points(new_point_data, 'time', 'x', 'y', 'precip', rem_sites1, to_crs, to_crs, method, digits, min_val).rename(columns={'precip': 'point_precip'})

        ## Extract and compare
        grid_points = interp2d.grid_to_points(grid_grid, 'time', 'x', 'y', 'precip', rem_sites1, to_crs, to_crs, 2, digits, min_val).rename(columns={'precip': 'grid_precip'})

        comp_data1 = pd.merge(rem_point_data, points_points.reset_index(), on=['time', 'x', 'y'])
        comp_data2 = pd.merge(comp_data1, grid_points.reset_index(), on=['time', 'x', 'y']).set_index(['time', 'x', 'y']).sort_index()

        return comp_data2
